;;;
;;; Universal Monitor Z8000
;;;   Copyright (C) 2020,2021 Haruo Asano
;;;

	CPU	Z8002

TARGET:	EQU	"Z8002"

	SUPMODE	ON
	
	INCLUDE	"config.inc"

	INCLUDE	"../common.inc"

;;;
;;; ROM area
;;;

	ORG	0000H

INIVEC:
	;; Reset
	DW	0
	DW	4000H		; FCW: System mode
	DW	CSTART
INIVECE:

	ORG	0100H
CSTART:	
	LDA	R15,STACK

	CALL	INIT

	LDA	R0,RAM_B
	LD	DSADDR,R0
	LD	GADDR,R0
	LD	SADDR,R0

	;; Opening message
	LDA	R1,OPNMSG
	CALL	STROUT

WSTART:
	LDA	R1,PROMPT
	CALL	STROUT
	CALL	GETLIN
	LDA	R1,INBUF
	CALL	SKIPSP
	CALL	UPPER
	TESTB	RL0
	JR	Z,WSTART

	CP	RL0,#'D'
	JP	Z,DUMP
	CP	RL0,#'G'
	JP	Z,GO
	CP	RL0,#'S'
	JP	Z,SETM

	CP	RL0,#'L'
	JP	Z,LOADH

	CP	RL0,#'I'
	JP	Z,PIN
	CP	RL0,#'O'
	JP	Z,POUT

ERR:
	LDA	R1,ERRMSG
	CALL	STROUT

	JR	WSTART

;;;
;;; DUMP memory
;;;

DUMP:
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	TEST	RL2
	JR	NZ,DP0
	;; No arg.
	CALL	SKIPSP
	TEST	RL0
	JR	NZ,ERR

	LD	R4,DSADDR
	LD	R5,R4
	ADD	R5,#128		; DEADDR
	JP	DPM

	;; 1st arg. found
DP0:
	LD	R4,R3		; DSADDR
	LD	RL0,@R1
	CP	RL0,#','
	JR	Z,DP1
	TEST	RL0
	JR	NZ,ERR
	;; No 2nd arg.
	LD	R5,R4
	ADD	R5,#128
	JP	DPM
DP1:
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	TEST	RL2
	JR	Z,ERR
	TESTB	@R1
	JR	NZ,ERR
	LD	R5,R3
	INC	R5
DPM:
	;; DUMP main
	LD	R6,R4
	AND	R6,#0FFF0H
	CLR	RH2		; DSTATE
DPM0:
	CALL	DPL
	CALL	CONST
	TEST	RL0
	JR	NZ,DPM1		; abort
	CP	RH2,#2
	JR	C,DPM0
	LD	DSADDR,R5
	JP	WSTART
DPM1:
	LD	DSADDR,R6
	CALL	CONIN
	JP	WSTART

DPL:
	;; DUMP line
	LD	R1,R6
	CALL	HEXOUT4
	LDA	R1,DSEP0
	CALL	STROUT
	LDA	R1,INBUF
	LD	RL2,#16
DPL0:
	CALL	DPB
	DBJNZ	RL2,DPL0

	LDA	R1,DSEP1
	CALL	STROUT

	LDA	R1,INBUF
	LD	RL2,#16
DPL1:
	LD	RL0,@R1
	INC	R1
	CP	RL0,#' '
	JR	C,DPL2
	CP	RL0,#7FH
	JR	NC,DPL2
	CALL	CONOUT
	JR	DPL3
DPL2:
	LD	RL0,#'.'
	CALL	CONOUT
DPL3:
	DBJNZ	RL2,DPL1
	JP	CRLF

DPB:
	;; DUMP byte
	LD	RL0,#' '
	CALL	CONOUT
	TEST	RH2
	JR	NZ,DPB2
	;; State 0
	CP	R6,R4
	JR	Z,DPB1
DPB0:
	;; Still 0 or 2
	LD	RL0,#' '
	LD	@R1,RL0
	CALL	CONOUT
	LD	RL0,#' '
	CALL	CONOUT
	INC	R6
	INC	R1
	RET
DPB1:
	;; Found start address
	LD	RH2,#1
DPB2:
	CP	RH2,#1
	JR	NZ,DPB0		; state 2
	;; DUMP state 1
	LD	RL0,@R6
	LD	@R1,RL0
	CALL	HEXOUT2
	INC	R6
	INC	R1

	CP	R6,R5
	JR	NZ,DPB3
	;; Found end address
	LD	RH2,#2
DPB3:
	RET

;;;
;;; GO address
;;;

GO:
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	TESTB	@R1
	JP	NZ,ERR
	TEST	RL2
	JR	Z,G0

	LD	GADDR,R3
G0:
	LD	R1,GADDR
	JP	@R1

;;;
;;; SET memory
;;;

SETM:	
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	TEST	RL2
	JR	Z,SM0
	LD	SADDR,R3
SM0:
	CALL	SKIPSP
	TEST	RL0
	JP	NZ,ERR
	LD	R6,SADDR
SM1:
	LD	R1,R6
	CALL	HEXOUT4
	LDA	R1,DSEP1
	CALL	STROUT
	LD	RL0,@R6
	CALL	HEXOUT2
	LD	RL0,#' '
	CALL	CONOUT
	CALL	GETLIN
	LDA	R1,INBUF
	CALL	SKIPSP
	TEST	RL0
	JR	NZ,SM2
	;; Empty (Increment address)
	INC	R6
	LD	SADDR,R6
	JR	SM1
SM2:
	CP	RL0,#'-'
	JR	NZ,SM3
	;; '-' (Decrement address)
	DEC	R6
	LD	SADDR,R6
	JR	SM1
SM3:
	CP	RL0,#'.'
	JR	NZ,SM4
	;; '.' (Quit)
	LD	SADDR,R6
	JP	WSTART
SM4:
	CALL	RDHEX
	TEST	RL2
	JP	Z,ERR
	LD	@R6,RL3
	INC	R6
	LD	SADDR,R6
	JR	SM1

;;;
;;; LOAD HEX file
;;;

LOADH:
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	TEST	RL0
	JP	NZ,ERR

	TEST	RL2
	JR	NZ,LH0

	CLR	R3		; Offset
LH0:
	CALL	CONIN
	CALL	UPPER
	CP	RL0,#'S'
	JR	Z,LHS0
LH1:
	CP	RL0,#':'
	JR	Z,LHI0
LH2:
	;; Skip to EOL
	CP	RL0,#CR
	JR	Z,LH0
	CP	RL0,#LF
	JR	Z,LH0
LH3:
	CALL	CONIN
	JR	LH2

LHI0:
	CALL	HEXIN
	LD	RH2,RL0		; Checksum
	LD	RL2,RL0		; Length

	CALL	HEXIN
	LD	RH1,RL0		; Address H
	ADD	RH2,RL0		; Checksum

	CALL	HEXIN
	LD	RL1,RL0		; Address L
	ADD	RH2,RL0		; Checksum

	;; Add offset
	ADD	R1,R3

	CALL	HEXIN
	LD	RL4,RL0		; RECTYP
	ADD	RH2,RL0		; Checksum

	TEST	RL2
	JR	Z,LHI3
LHI1:
	CALL	HEXIN
	ADD	RH2,RL0		; Checksum

	TEST	RL4
	JR	NZ,LHI2

	LD	@R1,RL0
	INC	R1
LHI2:
	DBJNZ	RL2,LHI1
LHI3:
	CALL	HEXIN
	ADD	RH2,RL0
	JR	NZ,LHIE		; Checksum error

	TEST	RL4
	JR	Z,LH3
	JP	WSTART
LHIE:
	LDA	R1,IHEMSG
	CALL	STROUT
	JP	WSTART

LHS0:
	CALL	CONIN
	LD	RL4,RL0		; RECTYP

	CALL	HEXIN
	LD	RL2,RL0		; Length+3
	LD	RH2,RL0		; Checksum

	CALL	HEXIN
	LD	RH1,RL0		; Address H
	ADD	RH2,RL0		; Checksum

	CALL	HEXIN
	LD	RL1,RL0		; Addreess L
	ADD	RH2,RL0		; Checksum

	ADD	R1,R3

	SUB	RL2,#3
	JR	Z,LHS3
LHS1:
	CALL	HEXIN
	ADD	RH2,RL0		; Checksum

	CP	RL4,#'1'
	JR	NZ,LHS2

	LD	@R1,RL0
	INC	R1
LHS2:
	DBJNZ	RL2,LHS1
LHS3:
	CALL	HEXIN
	ADD	RH2,RL0
	CP	RH2,#0FFH
	JR	NZ,LHSE

	CP	RL4,#'9'
	JP	Z,WSTART
	JR	LH3
LHSE:
	LDA	R1,SHEMSG
	CALL	STROUT
	JP	WSTART

;;;
;;; Port in
;;;

PIN:
	CLR	R5
PI0:
	INC	R1
	LD	RL0,@R1
	CALL	UPPER
	CP	RL0,#'S'
	JR	NZ,PI1
	OR	R5,#02H
	JR	PI0
PI1:
	CP	RL0,#'W'
	JR	NZ,PI2
	OR	R5,#04H
	JR	PI0
PI2:
	CALL	SKIPSP
	CALL	RDHEX
	TEST	RL2
	JP	Z,ERR		; Port no. missing
	CALL	SKIPSP
	TEST	RL0
	JP	NZ,ERR

	LD	R1,PITAB(R5)
	CALL	@R1
	CALL	CRLF
	JP	WSTART

PITAB:
	DW	PINB
	DW	PISB
	DW	PINW
	DW	PISW

PINB:	
	;; Normal - Byte
	INB	RL0,@R3
	JP	HEXOUT2

PISB:
	;; Special - Byte
	LD	INBUF,#3A85H	; SINB RL0,xxxx
	LD	INBUF+2,R3
	LD	INBUF+4,#9E08H	; RET
	CALL	INBUF
	JP	HEXOUT2

PINW:
	;; Normal - Word
	IN	R1,@R3
	JP	HEXOUT4

PISW:
	;; Special - Word
	LD	INBUF,#3B15H	; SIN R1,xxxx
	LD	INBUF+2,R3
	LD	INBUF+4,#9E08H	; RET
	CALL	INBUF
	JP	HEXOUT4
	
;;;
;;; Port out
;;;

POUT:
	CLR	RL5
PO0:
	INC	R1
	LD	RL0,@R1
	CALL	UPPER
	CP	RL0,#'S'
	JR	NZ,PO1
	OR	RL5,#02H
	JR	PO0
PO1:
	CP	RL0,#'W'
	JR	NZ,PO2
	OR	RL5,#04H
	JR	PO0
PO2:
	CALL	SKIPSP
	CALL	RDHEX
	TEST	RL2
	JP	Z,ERR		; Port no. missing
	LD	R4,R3
	CALL	SKIPSP
	CP	RL0,#','
	JP	NZ,ERR
	INC	R1
	CALL	SKIPSP
	CALL	RDHEX
	TEST	RL2
	JP	Z,ERR		; Data missing
	CALL	SKIPSP
	TEST	RL0
	JP	NZ,ERR
	
	LD	R1,POTAB(R5)
	CALL	@R1
	JP	WSTART

POTAB:
	DW	PONB
	DW	POSB
	DW	PONW
	DW	POSW

PONB:
	;; Normal - Byte
	OUTB	@R4,RL3
	RET

POSB:
	;; Special - Byte
	LD	INBUF,#3AB7H	; SOUTB xxxx,RL3
	LD	INBUF+2,R4
	LD	INBUF+4,#9E08H	; RET
	CALL	INBUF
	RET

PONW:
	;; Normal - Word
	OUT	@R4,R3
	RET

POSW:
	;; Special - Word
	LD	INBUF,#3B37H	; SOUT xxxx,R3
	LD	INBUF+2,R4
	LD	INBUF+4,#9E08H	; RET
	CALL	INBUF
	RET

;;;
;;; Other support routines
;;;

STROUT:
	LDB	RL0,@R1
	TESTB	RL0
	JR	Z,STROE
	CALL	CONOUT
	INC	R1
	JR	STROUT
STROE:
	RET

HEXOUT4:
	LD	RL0,RH1
	CALL	HEXOUT2
	LD	RL0,RL1
HEXOUT2:
	PUSH	@R15,R0
	SRL	RL0,#4
	CALL	HEXOUT1
	POP	R0,@R15
HEXOUT1:
	AND	RL0,#0FH
	ADD	RL0,#'0'
	CP	RL0,#'9'+1
	JP	C,CONOUT
	ADD	RL0,#'A'-'9'-1
	JP	CONOUT

HEXIN:
	CLR	RL0
	CALL	HI0
	SLL	RL0,#4
HI0:
	PUSH	@R15,R1
	LD	RL1,RL0
	CALL	CONIN
	CALL	UPPER
	CP	RL0,#'0'
	JR	C,HIR
	CP	RL0,#'9'+1
	JR	C,HI1
	CP	RL0,#'A'
	JR	C,HIR
	CP	RL0,#'F'+1
	JR	NC,HIR
	SUB	RL0,#'A'-'9'-1
HI1:
	SUB	RL0,#'0'
	OR	RL0,RL1
HIR:
	POP	R1,@R15
	RET

CRLF:
	LD	RL0,#CR
	CALL	CONOUT
	LD	RL0,#LF
	JP	CONOUT

GETLIN:
	LDA	R1,INBUF
	CLR	RL2
GL0:
	CALL	CONIN
	CP	RL0,#CR
	JR	Z,GLE
	CP	RL0,#LF
	JR	Z,GLE
	CP	RL0,#BS
	JR	Z,GLB
	CP	RL0,#DEL
	JR	Z,GLB
	CP	RL0,#' '
	JR	C,GL0
	CP	RL0,#80H
	JR	NC,GL0
	CP	RL2,#BUFLEN-1
	JR	NC,GL0		; Too long
	LD	@R1,RL0
	CALL	CONOUT
	INC	R1
	INC	RL2
	JR	GL0
GLB:
	TEST	RL2
	JR	Z,GL0
	DEC	R1
	DEC	RL2
	LD	RL0,#BS
	CALL	CONOUT
	LD	RL0,#' '
	CALL	CONOUT
	LD	RL0,#BS
	CALL	CONOUT
	JR	GL0
GLE:
	LDB	@R1,#00H
	JP	CRLF

SKIPSP:
	LD	RL0,@R1
	CP	RL0,#' '
	JR	NZ,SSR
	INC	R1
	JR	SKIPSP
SSR:
	RET

UPPER:
	CP	RL0,#'a'
	JR	C,UPR
	CP	RL0,#'z'+1
	JR	NC,UPR
	ADD	RL0,#'A'-'a'
UPR:
	RET

RDHEX:
	CLR	RL2
	CLR	R3
RDH0:
	LD	RL0,@R1
	CALL	UPPER
	CP	RL0,#'0'
	JR	C,RDHE
	CP	RL0,#'9'+1
	JR	C,RDH1
	CP	RL0,#'A'
	JR	C,RDHE
	CP	RL0,#'F'+1
	JR	NC,RDHE
	SUB	RL0,#'A'-'9'-1
RDH1:
	SUB	RL0,#'0'
	SLA	R3,4
	OR	RL3,RL0
	INC	R1
	INC	RL2
	JR	RDH0
RDHE:
	RET


OPNMSG:
	DB	CR,LF,"Universal Monitor Z8000",CR,LF,00H

PROMPT:
	DB	"] ",00H

IHEMSG:
	DB	"Error ihex",CR,LF,00H
SHEMSG:
	DB	"Error srec",CR,LF,00H
ERRMSG:
	DB	"Error",CR,LF,00H

DSEP0:
	DB	" :",00H
DSEP1:
	DB	" : ",00H

	ALIGN 2
	
	IF USE_DEV_8530
	INCLUDE	"dev/dev_8530.asm"
	ENDIF

;;;
;;; RAM area
;;;

	;;
	;; Work area
	;;

	ORG	WORK_B

INBUF:	DS	BUFLEN		; Line input buffer
DSADDR:	DS	2		; DUMP start address
SADDR:	DS	2		; SET address
GADDR:	DS	2		; GO address

	END
